""" See The Python Standard Library, Runtime Services, __main__.

    This __main__.py module is included to execute code in
    the package when it is run with python -m phmdoctest.
"""

from .main import entry_point
entry_point()
